import hashlib
import logging
from base64 import b64encode
from datetime import datetime
from pathlib import Path
import validators
import asn1crypto
from ssl import PEM_cert_to_DER_cert
from OpenSSL.crypto import X509, X509Name, dump_certificate, load_certificate, FILETYPE_PEM, FILETYPE_ASN1, FILETYPE_TEXT, TYPE_RSA, TYPE_DSA, TYPE_DH, TYPE_EC
from cryptography.x509 import Certificate
from certvalidator.errors import PathValidationError, RevokedError, InvalidCertificateError, PathBuildingError
from validators.utils import ValidationFailure
from tlsverify import util
from tlsverify.exceptions import ValidationError
from tabulate import tabulate

__module__ = 'tlsverify'

logger = logging.getLogger(__name__)

class Validator:
    _pem :bytes
    _der :bytes
    x509 :X509
    certificate :Certificate
    _pem_certificate_chain :list
    certificate_chain :list
    metadata :util.Metadata

    def __init__(self, host :str = None, port :int = 443) -> None:
        self.certificate_valid = False
        self.validation_checks = {}
        self.certificate_verify_messages = []
        self.certificate_chain_valid = None
        self.certificate_chain_validation_result = None
        self._pem = None
        self._der = None
        self.x509 = None
        self.certificate = None
        self._pem_certificate_chain = []
        self.certificate_chain = []
        self.metadata = None
        if host is not None:
            self.x509, self.certificate_chain, valid_subjects, protocol, cipher = util.get_certificates(host, port) #TODO handle clientAuth
            logger.debug('added pyOpenSSL x509')
            self.metadata = util.Metadata(host=host, port=port, negotiated_cipher=cipher, negotiated_protocol=protocol)
            if self.x509 is None or not self.certificate_chain:
                raise ValidationError(f'Unable to negotiate a TLS socket connection with server at {host}:{port} to obtain the Certificate')
            self._pem = dump_certificate(FILETYPE_PEM, self.x509)
            logger.debug('added PEM bytes')
            self._der = PEM_cert_to_DER_cert(self._pem.decode())
            logger.debug('added ASN1/DER bytes')
            self.certificate = self.x509.to_cryptography()
            logger.debug('added lib cryptography object')
            self._pem_certificate_chain = Validator.convert_x509_to_PEM(self.certificate_chain)

    def cert_to_text(self) -> str:
        logger.debug('dump_certificate x509 FILETYPE_TEXT')
        return dump_certificate(FILETYPE_TEXT, self.x509).decode()

    def tabulate(self) -> str:
        def any_to_string(value, delimiter='\n') -> str:
            if isinstance(value, str):
                return value
            if isinstance(value, list) and isinstance(value[0], str):
                return delimiter.join(value)
            if isinstance(value, list) and not isinstance(value[0], str):
                n = []
                for d in value:
                    n.append(any_to_string(d, delimiter))
                return any_to_string(n)
            if isinstance(value, dict):
                return delimiter.join([f'{key}={str(value[key])}' for key in value.keys()])
            return str(value)
        skip = ['certificate_san', 'certificate_extensions', 'subjectKeyIdentifier', 'authorityKeyIdentifier']
        kv = [
            ['certificate_valid', self.certificate_valid],
            ['certificate_chain_valid', self.certificate_chain_valid],
            ['certificate_chain_validation_result', self.certificate_chain_validation_result]
        ]
        kv += [['Error', err] for err in self.certificate_verify_messages]
        kv += [[f'Check {key}', self.validation_checks[key]] for key in self.validation_checks.keys()]
        kv += [[key, getattr(self.metadata, key)] for key in list(vars(self.metadata).keys()) if key not in skip]
        kv += [[v['name'], any_to_string(v, ' ') if v['name'] not in v else any_to_string(v[v['name']], ' ')] for v in self.metadata.certificate_extensions if v['name'] not in skip]
        return tabulate(kv, tablefmt='tsv', disable_numparse=True, colalign=("right",))

    def init_der(self, der :bytes):
        self._der = der
        logger.debug('added ASN1/DER bytes')
        self.x509 = load_certificate(FILETYPE_ASN1, der)
        logger.debug('added pyOpenSSL x509')
        self._pem = dump_certificate(FILETYPE_PEM, self.x509)
        logger.debug('added PEM bytes')
        self.certificate = self.x509.to_cryptography()
        logger.debug('added lib cryptography object')

    def init_pem(self, pem :bytes):
        self._pem = pem
        logger.debug('added PEM bytes')
        self.x509 = load_certificate(FILETYPE_PEM, pem)
        logger.debug('added pyOpenSSL x509')
        self._der = PEM_cert_to_DER_cert(self._pem.decode())
        logger.debug('added ASN1/DER bytes')
        self.certificate = self.x509.to_cryptography()
        logger.debug('added lib cryptography object')

    def init_x509(self, x509 :X509):
        self.x509 = x509
        logger.debug('added pyOpenSSL x509')
        self._pem = dump_certificate(FILETYPE_PEM, x509)
        logger.debug('added PEM bytes')
        self._der = PEM_cert_to_DER_cert(self._pem.decode())
        logger.debug('added ASN1/DER bytes')
        self.certificate = x509.to_cryptography()
        logger.debug('added lib cryptography object')

    @staticmethod
    def convert_decimal_to_serial_bytes(decimal :int):
        # add leading 0
        a = "0%x" % decimal
        # force even num bytes, remove leading 0 if necessary
        b = a[1:] if len(a)%2==1 else a
        return format(':'.join(s.encode('utf8').hex().lower() for s in b))

    @staticmethod
    def str_n_split(input :str, n :int = 2, delimiter :str = ' '):
        return delimiter.join([input[i:i+n] for i in range(0, len(input), n)])

    @staticmethod
    def convert_x509_to_PEM(certificate_chain :list) -> list[bytes]:
        pem_certs = []
        for cert in certificate_chain:
            if not isinstance(cert, X509):
                raise ValidationError(f'convert_x509_to_PEM expected OpenSSL.crypto.X509, got {type(cert)}')
            pem_certs.append(dump_certificate(FILETYPE_PEM, cert))
        return pem_certs

    def extract_metadata(self):
        if not hasattr(self, 'metadata') or self.metadata is None:
            self.metadata = util.Metadata(host=None)
        public_key = self.x509.get_pubkey()
        if public_key.type() == TYPE_RSA:
            self.metadata.certificate_public_key_type = 'RSA'
        if public_key.type() == TYPE_DSA:
            self.metadata.certificate_public_key_type = 'DSA'
        if public_key.type() == TYPE_DH:
            self.metadata.certificate_public_key_type = 'DH'
        if public_key.type() == TYPE_EC:
            self.metadata.certificate_public_key_type = 'EC'
        self.metadata.certificate_key_size = public_key.bits()
        self.metadata.certificate_serial_number_decimal = self.x509.get_serial_number()
        self.metadata.certificate_serial_number = Validator.convert_decimal_to_serial_bytes(self.x509.get_serial_number())
        self.metadata.certificate_serial_number_hex = '{0:#0{1}x}'.format(self.x509.get_serial_number(), 4)
        subject = self.x509.get_subject()
        self.metadata.certificate_subject = "".join("/{0:s}={1:s}".format(name.decode(), value.decode()) for name, value in subject.get_components())
        issuer: X509Name = self.x509.get_issuer()
        self.metadata.certificate_issuer = issuer.commonName
        self.metadata.certificate_issuer_country = issuer.countryName
        self.metadata.certificate_signature_algorithm = self.x509.get_signature_algorithm().decode('ascii')
        self.metadata.certificate_pin_sha256 = b64encode(hashlib.sha256(asn1crypto.x509.Certificate.load(self._der).public_key.dump()).digest()).decode()
        self.metadata.certificate_sha256_fingerprint = hashlib.sha256(self._der).hexdigest()
        self.metadata.certificate_sha1_fingerprint = hashlib.sha1(self._der).hexdigest()
        self.metadata.certificate_md5_fingerprint = hashlib.md5(self._der).hexdigest()
        self.metadata.certificate_san = util.get_san(self.certificate)
        self.metadata.certificate_valid_tls_usage = util.key_usage_exists(self.certificate, 'digital_signature') is True and util.key_usage_exists(self.certificate, 'serverAuth') is True
        not_before = datetime.strptime(self.x509.get_notBefore().decode('ascii'), util.X509_DATE_FMT)
        not_after = datetime.strptime(self.x509.get_notAfter().decode('ascii'), util.X509_DATE_FMT)
        self.metadata.certificate_not_before = not_before.isoformat()
        self.metadata.certificate_not_after = not_after.isoformat()
        self.metadata.certificate_common_name = util.extract_certificate_common_name(self.certificate)
        for ext in self.metadata.certificate_extensions:
            if ext['name'] == 'TLSFeature' and 'rfc6066' in ext['features']:
                self.metadata.revocation_ocsp_stapling = True
                self.metadata.revocation_ocsp_must_staple = True
            if ext['name'] == 'subjectKeyIdentifier':
                self.metadata.certificate_subject_key_identifier = ext[ext['name']]
            if ext['name'] == 'authorityKeyIdentifier':
                self.metadata.certificate_authority_key_identifier = ext[ext['name']]

    def client_authentication(self, valid_subjects :list[tuple], client_pem :list[str], client_ca :str = None, cafiles :list[str] = None) -> bool:
        """
        valid_subjects: [[(b'C', b'US'),(b'ST', b'California'),(b'L', b'San Francisco'),(b'O', b'BadSSL'),(b'CN', b'BadSSL Client Root Certificate Authority')]]
        """
        raise NotImplementedError()

    def verify(self, host :str = None, port :int = None, peer :bool = False) -> bool:
        if not hasattr(self, 'metadata') or self.metadata is None:
            self.metadata = util.Metadata(host=host, port=port)
        if peer is False:
            logger.debug('Server certificate validations')
            if isinstance(host, str):
                self.metadata.host = host
            if validators.domain(self.metadata.host) is not True:
                raise ValueError(f"provided an invalid domain {host}")
            if isinstance(port, int):
                self.metadata.port = port
            self.validation_checks['certificate_valid_tls_usage'] = util.key_usage_exists(self.certificate, 'digital_signature') is True and util.key_usage_exists(self.certificate, 'serverAuth') is True
            self.validation_checks['common_name_valid'] = util.validate_common_name(self.metadata.certificate_common_name, self.metadata.host) is True
            self.validation_checks['match_hostname'] = util.match_hostname(self.metadata.host, self.certificate)
            self.metadata.certificate_is_self_signed = util.is_self_signed(self.certificate)
            self.validation_checks['not_self_signed'] = self.metadata.certificate_is_self_signed is False
            if self.metadata.certificate_is_self_signed:
                self.validation_checks['trusted_ca'] = False
                self.certificate_verify_messages.append('The CA is not properly imported as a trusted CA into the browser, Chrome based browsers will block visitors and show them ERR_CERT_AUTHORITY_INVALID')
        logger.debug('Common certificate validations')
        not_after = datetime.fromisoformat(self.metadata.certificate_not_after)
        not_before = datetime.fromisoformat(self.metadata.certificate_not_before)
        self.validation_checks['not_expired'] = not_after > datetime.utcnow()
        if self.validation_checks['not_expired'] is False:
            self.certificate_verify_messages.append(f'Expired {(datetime.utcnow() - not_after).days} days ago' if not_after < datetime.utcnow() else f'Valid for {(not_after - datetime.utcnow()).days} days')
        self.validation_checks['issued_past_tense'] = not_before < datetime.utcnow()
        if self.validation_checks['issued_past_tense'] is False:
            self.certificate_verify_messages.append(f'Will only be valid for use in {(datetime.utcnow() - self.metadata.certificate_not_before).days} days')
        self.validation_checks['common_name_defined'] = self.metadata.certificate_common_name is not None
        self.validation_checks['avoid_known_weak_signature_algorithm'] = self.metadata.certificate_signature_algorithm not in util.KNOWN_WEAK_SIGNATURE_ALGORITHMS.keys()
        if self.validation_checks['avoid_known_weak_signature_algorithm'] is False:
            self.certificate_verify_messages.append(util.KNOWN_WEAK_SIGNATURE_ALGORITHMS[self.metadata.certificate_signature_algorithm])
        self.validation_checks['avoid_known_weak_keys'] = self.metadata.certificate_public_key_type not in util.KNOWN_WEAK_KEYS.keys() or self.metadata.certificate_key_size > util.WEAK_KEY_SIZE[self.metadata.certificate_public_key_type]
        if self.validation_checks['avoid_known_weak_keys'] is False:
            self.certificate_verify_messages.append(util.KNOWN_WEAK_KEYS[self.metadata.certificate_public_key_type])
        self.validation_checks['avoid_deprecated_protocols'] = self.metadata.negotiated_protocol not in util.WEAK_PROTOCOL.keys()
        if self.validation_checks['avoid_deprecated_protocols'] is False:
            self.certificate_verify_messages.append(util.WEAK_PROTOCOL[self.metadata.negotiated_protocol])
        self.certificate_valid = all(list(self.validation_checks.values()))
        return self.certificate_valid

    def verify_chain(self, certificate_chain :list[bytes] = None) -> bool:
        if certificate_chain is not None:
            self._pem_certificate_chain = certificate_chain
        self.metadata.certificate_extensions, validator_key_usage, validator_extended_key_usage = util.gather_key_usages(self.certificate)
        self.validation_checks['not_revoked'] = True
        try:
            logger.debug('certificate chain validation')
            util.validate_certificate_chain(self._der, self._pem_certificate_chain, validator_key_usage, validator_extended_key_usage)
        except RevokedError as ex:
            logger.debug(ex, exc_info=True)
            self.validation_checks['not_revoked'] = False
            self.certificate_chain_valid = False
            self.certificate_chain_validation_result = str(ex)
        except InvalidCertificateError as ex:
            logger.debug(ex, exc_info=True)
            self.validation_checks['certificate_chain_valid'] = False
            self.certificate_chain_valid = False
            self.certificate_chain_validation_result = str(ex)
        except PathValidationError as ex:
            logger.debug(ex, exc_info=True)
            self.validation_checks['certificate_chain_valid'] = False
            self.certificate_chain_valid = False
            self.certificate_chain_validation_result = str(ex)
        except PathBuildingError as ex:
            logger.debug(ex, exc_info=True)
            self.validation_checks['certificate_chain_valid'] = False
            self.certificate_chain_valid = False
            self.certificate_chain_validation_result = str(ex)
        except Exception as ex:
            logger.warning(ex, exc_info=True)
            self.validation_checks['certificate_chain_valid'] = False
            self.certificate_chain_validation_result = str(ex)

        if self.certificate_chain_validation_result is None:
            self.certificate_chain_valid = True
            self.certificate_chain_validation_result = f'Validated: {",".join(validator_key_usage + validator_extended_key_usage)}'

        self.certificate_valid = all(list(self.validation_checks.values()))
        self.extract_metadata()
        return self.certificate_valid and self.certificate_chain_valid

def verify(host :str, port :int = 443, cafiles :list = None, tlsext :bool = False, client_pem :str = None, client_ca :str = None) -> tuple[bool,list[Validator]]:
    additional_cert = None
    peer_validations = []
    logger.info('Testing TLS connection')
    x509, x509_certificate_chain, valid_subjects, protocol, cipher = util.get_certificates(host, port, cafiles=cafiles, client_pem=client_pem, client_ca=client_ca, tlsext=tlsext)
    sni_support = None
    if isinstance(x509, X509) and tlsext is True:
        logger.info('SNI supported')
        sni_support = True
    if not isinstance(x509, X509) and tlsext is False:
        logger.info('SNI not support')
        x509, x509_certificate_chain, extra_valid_subjects, protocol, cipher = util.get_certificates(host, port, cafiles=cafiles, client_pem=client_pem, client_ca=client_ca, tlsext=False)
        sni_support = False
        valid_subjects += extra_valid_subjects
    if not isinstance(x509, X509):
        raise ValidationError('Unable to negotiate a tls connection')
    if isinstance(x509, X509) and tlsext is False:
        logger.info('Checking SNI support')
        additional_cert, _, additional_valid_subjects, _, _ = util.get_certificates(host, port, cafiles=cafiles, client_pem=client_pem, client_ca=client_ca, tlsext=True)
        sni_support = isinstance(additional_cert, X509)
        valid_subjects += additional_valid_subjects
        
    logger.info('Checking server certificate')
    validator = Validator()
    validator.init_x509(x509)
    ca, path_length = util.get_basic_constraints(validator.certificate)
    if isinstance(ca, bool) and ca is True:
        raise ValidationFailure('server cert should not be a CA')
    validator.extract_metadata()
    validator.metadata.negotiated_cipher = cipher
    validator.metadata.negotiated_protocol = protocol
    validator.metadata.sni_support = sni_support
    if isinstance(path_length, int):
        validator.metadata.tlsext_basic_constraints_path_length = len(x509_certificate_chain) == path_length
    validator.verify(host, port)
    if len(valid_subjects) > 0 and util.key_usage_exists(validator.certificate, 'clientAuth') is True:
        pass
        # validator.client_authentication(valid_subjects, cafiles=cafiles, client_pem=client_pem, client_ca=client_ca)

    for cert in x509_certificate_chain:
        logger.info('Checking peer certificate')
        peer_validator = Validator()
        peer_validator.init_x509(cert)
        _, peer_path_length = util.get_basic_constraints(peer_validator.certificate)
        peer_validator.extract_metadata()
        if isinstance(peer_path_length, int):
            peer_validator.metadata.tlsext_basic_constraints_path_length = len(x509_certificate_chain) == peer_path_length
        peer_validator.verify(peer=True)
        peer_validations.append(peer_validator)
    validator.verify_chain(Validator.convert_x509_to_PEM(x509_certificate_chain))
    validations = peer_validations
    if isinstance(additional_cert, X509):
        additional_validator = Validator()
        additional_validator.init_x509(additional_cert)
        additional_validator.extract_metadata()
        if validator.metadata.certificate_common_name != additional_validator.metadata.certificate_common_name:
            logger.info('Checking additional SNI negotiated certificate')
            additional_validator.verify(host, port)
            tmp_valid = all([v.certificate_valid for v in peer_validations + [validator]] + [additional_validator.certificate_valid])
            if tmp_valid is False:
                logger.info('Checking peer certificate')
                additional_validator.verify_chain(Validator.convert_x509_to_PEM(x509_certificate_chain))
            validations.append(additional_validator)
        if len(valid_subjects) > 0 and util.key_usage_exists(additional_validator.certificate, 'clientAuth') is True:
            pass
            # additional_validator.client_authentication(valid_subjects, cafiles=cafiles, client_pem=client_pem, client_ca=client_ca)
    validations.append(validator)

    valid = all([v.certificate_valid for v in validations])
    return valid, validations



import logging
import argparse
from datetime import datetime
from . import verify

__module__ = 'tlsverify.cli'

def cli():
    evaluation_start = datetime.utcnow()
    parser = argparse.ArgumentParser()
    parser.add_argument('-H', '--host', help='host to check', dest='host', required=True)
    parser.add_argument('-p', '--port', help='TLS port of host', dest='port', default=443)
    parser.add_argument('-c', '--cafiles', help='path to PEM encoded CA bundle file, url or file path accepted', dest='cafiles', default=None)
    parser.add_argument('-C', '--client-pem', help='path to PEM encoded client certificate, url or file path accepted', dest='client_pem', default=None)
    parser.add_argument('-T', '--client-ca-pem', help='path to PEM encoded client CA certificate, url or file path accepted', dest='client_ca', default=None)
    parser.add_argument('--sni', help='Negotiate SNI via PyOpenSSL Connection set_tlsext_host_name and INDA encoded host', dest='tlsext', action="store_true")
    parser.add_argument('-v', '--errors-only', help='set logging level to ERROR (default CRITICAL)', dest='log_level_error', action="store_true")
    parser.add_argument('-vv', '--warning', help='set logging level to WARNING (default CRITICAL)', dest='log_level_warning', action="store_true")
    parser.add_argument('-vvv', '--info', help='set logging level to INFO (default CRITICAL)', dest='log_level_info', action="store_true")
    parser.add_argument('-vvvv', '--debug', help='set logging level to DEBUG (default CRITICAL)', dest='log_level_debug', action="store_true")
    args = parser.parse_args()
    log_level = logging.CRITICAL
    if args.log_level_error:
        log_level = logging.ERROR
    if args.log_level_warning:
        log_level = logging.WARNING
    if args.log_level_info:
        log_level = logging.INFO
    if args.log_level_debug:
        log_level = logging.DEBUG
    logging.basicConfig(
        format='%(asctime)s - %(name)s - [%(levelname)s] %(message)s',
        level=log_level
    )
    valid, results = verify(args.host, int(args.port), cafiles=args.cafiles, tlsext=args.tlsext, client_pem=args.client_pem, client_ca=args.client_ca)
    for validator in results:
        print(validator.tabulate())
        print('\n\n')
    print('\nValid ✓✓✓' if valid else '\nNot Valid. There where validation errors')
    print(f'evaluation duration seconds {(datetime.utcnow() - evaluation_start).total_seconds()}')



import logging
import string
import random
from urllib.request import urlretrieve
from dataclasses import dataclass, field
from binascii import hexlify
from socket import socket, AF_INET, SOCK_STREAM
from pathlib import Path
from cryptography import x509
from cryptography.x509 import Certificate, extensions, SubjectAlternativeName, DNSName
from OpenSSL import SSL
from OpenSSL.crypto import FILETYPE_PEM
from certifi import where
from certvalidator import CertificateValidator, ValidationContext
import validators
import idna

__module__ = 'tlsverify.util'

logger = logging.getLogger(__name__)
X509_DATE_FMT = r'%Y%m%d%H%M%SZ'
WEAK_KEY_SIZE = {
    'RSA': 1024,
    'DSA': 2048,
    'EC': 160,
}
KNOWN_WEAK_KEYS = {
    'RSA': '2000: Factorization of a 512-bit RSA Modulus, essentially derive a private key knowing only the public key. Verified bt EFF in 2001. Later in 2009 factorization of up to 1024-bit keys',
    'DSA': '1999: HPL Laboratories demonstrated lattice attacks on DSA, a non-trivial example of the known message attack that is a total break and message forgery technique. 2010 Dimitrios Poulakis demonstrated a lattice reduction technique for single or multiple message forgery',
    'EC': '2010 Dimitrios Poulakis demonstrated a lattice reduction technique to attack ECDSA for single or multiple message forgery',
}
KNOWN_WEAK_SIGNATURE_ALGORITHMS = {
    'sha1WithRSAEncryption': 'Macquarie University Australia 2009: identified vulnerabilities to collision attacks, later in 2017 Marc Stevens demonstrated collision proofs',
    'md5WithRSAEncryption': 'Arjen Lenstra and Benne de Weger 2005: vulnerable to hash collision attacks',
    'md2WithRSAEncryption': 'Rogier, N. and Chauvaud, P. in 1995: vulnerable to collision, later preimage resistance, and second-preimage resistance attacks were demonstrated at BlackHat 2008 by Mark Twain',
}
OPENSSL_VERSION_LOOKUP = {
    768: 'SSLv3',
    769: 'TLSv1',
    770: 'TLSv1.1',
    771: 'TLSv1.2',
    772: 'TLSv1.3',
}
WEAK_PROTOCOL = {
    'SSLv2': 'SSLv2 Deprecated in 2011 (rfc6176) with undetectable manipulator-in-the-middle exploits',
    'SSLv3': 'SSLv3 Deprecated in 2015 (rfc7568) mainly due to POODLE, a manipulator-in-the-middle exploit',
    'TLSv1': 'TLSv1 2018 deprecated by PCI Council. Also in 2018, Apple, Google, Microsoft, and Mozilla jointly announced deprecation. Officially deprecated in 2020 (rfc8996)',
    'TLSv1.1': 'TLSv1.1 No longer supported by Firefox 24 or newer and Chrome 29 or newer. Deprecated in 2020 (rfc8996)',
}

@dataclass
class Metadata:
    host :str = field(default_factory=str)
    certificate_public_key_type :str = field(default_factory=str)
    certificate_key_size :int = field(default_factory=int)
    certificate_serial_number :str = field(default_factory=str)
    certificate_serial_number_decimal :int = field(default_factory=str)
    certificate_serial_number_hex :str = field(default_factory=str)
    certificate_subject :str = field(default_factory=str)
    certificate_issuer :str = field(default_factory=str)
    certificate_issuer_country :str = field(default_factory=str)
    certificate_signature_algorithm :str = field(default_factory=str)
    certificate_pin_sha256 :str = field(default_factory=str)
    certificate_sha256_fingerprint :str = field(default_factory=str)
    certificate_sha1_fingerprint :str = field(default_factory=str)
    certificate_md5_fingerprint :str = field(default_factory=str)
    certificate_not_before :str = field(default_factory=str)
    certificate_not_after :str = field(default_factory=str)
    certificate_common_name :str = field(default_factory=str)
    certificate_san :list = field(default_factory=list)
    certificate_subject_key_identifier :str = field(default_factory=str)
    certificate_authority_key_identifier :str = field(default_factory=str)
    certificate_extensions :list = field(default_factory=list)
    certificate_is_self_signed : bool = field(default_factory=bool)
    tlsext_basic_constraints_path_length :int = field(default_factory=str)
    negotiated_cipher :str = field(default_factory=str)
    negotiated_protocol :str = field(default_factory=str)
    sni_support :bool = field(default_factory=bool)
    revocation_ocsp_stapling :bool = field(default_factory=bool)
    revocation_ocsp_must_staple :bool = field(default_factory=bool)
    port :int = 443

def filter_valid_files_urls(inputs :list[str], exception :Exception = None, tmp_path_prefix :str = '/tmp') -> list[str]:
    ret = set()
    for test in inputs:
        if test is None:
            raise exception
        file_path = Path(test)
        if file_path.is_file() is False and validators.url(test) is not True:
            raise exception
        if file_path.is_file():
            ret.add(test)
            continue
        if validators.url(test) is True:
            r = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(16))
            local_path = f'{tmp_path_prefix}/{r}.pem'
            urlretrieve(test, local_path)
            file_path = Path(local_path)
            if not file_path.is_file():
                raise exception
            ret.add(local_path)
    return list(ret)

def get_certificates(host :str, port :int = 443, cafiles :list = None, client_pem :str = None, client_ca :str = None, tlsext :bool = False, tmp_path_prefix :str = '/tmp') -> tuple[bytes,list,Metadata]:
    if not isinstance(port, int):
        raise TypeError(f"provided an invalid type {type(port)} for port, expected int")
    if validators.domain(host) is not True:
        raise ValueError(f"provided an invalid domain {host}")
    if not isinstance(tmp_path_prefix, str):
        raise TypeError(f"provided an invalid type {type(tmp_path_prefix)} for tmp_path_prefix, expected str")

    client_pem_err = AttributeError(f'client_pem was provided "{client_pem}" but is not a valid URL or file does not exist')
    if client_pem is not None:
        if not isinstance(client_pem, str):
            raise client_pem_err
        res = filter_valid_files_urls([client_pem], client_pem_err, tmp_path_prefix)
        if len(res) == 1:
            client_pem = res[0]
    client_ca_err = AttributeError(f'client_ca was provided "{client_ca}" but is not a valid URL or file does not exist')
    if client_ca is not None:
        if not isinstance(client_ca, str):
            raise client_ca_err
        res = filter_valid_files_urls([client_ca], client_ca_err, tmp_path_prefix)
        if len(res) == 1:
            client_ca = res[0]

    cafiles_err = AttributeError(f'cafiles was provided but is not a valid URLs or files do not exist\n{cafiles}')
    if cafiles is not None:
        if not isinstance(cafiles, list):
            raise cafiles_err
        cafiles = filter_valid_files_urls(cafiles, cafiles_err, tmp_path_prefix)
    negotiated_cipher = None
    negotiated_protocol = None
    x509 = None
    client_ca_list = []
    certificate_chain = []
    for version in [SSL.SSL3_VERSION, SSL.TLS1_VERSION, SSL.TLS1_1_VERSION, SSL.TLS1_2_VERSION, SSL.TLS1_3_VERSION]:
        logger.info(f'Trying protocol {OPENSSL_VERSION_LOOKUP[version]}')
        certificate_chain = []
        ctx = SSL.Context(method=SSL.SSLv23_METHOD)
        ctx.load_verify_locations(cafile=where())
        for cafile in cafiles or []:
            ctx.load_verify_locations(cafile=cafile)
        if client_pem is not None:
            ctx.use_certificate_file(certfile=client_pem, filetype=FILETYPE_PEM)
        if client_ca is not None:
            ctx.load_client_ca(cafile=client_ca)
        ctx.set_max_proto_version(version)
        ctx.check_hostname = False
        ctx.verify_mode = SSL.VERIFY_NONE
        conn = SSL.Connection(ctx, socket(AF_INET, SOCK_STREAM))
        conn.connect((host, port))
        conn.settimeout(3)
        if tlsext is True:
            logger.info('using SNI')
            conn.set_tlsext_host_name(idna.encode(host))
        conn.setblocking(1)
        try:
            conn.do_handshake()
            client_ca_list = conn.get_client_ca_list()
            x509 = conn.get_peer_certificate()
            negotiated_cipher = conn.get_cipher_name()
            negotiated_protocol = conn.get_protocol_version_name()
            for (_, cert) in enumerate(conn.get_peer_cert_chain()):
                certificate_chain.append(cert)
            logger.debug(f'Peer cert chain length: {len(certificate_chain)}')
        except SSL.Error as err:
            if 'no protocols available' not in str(err) and 'alert protocol' not in str(err):
                logger.warning(err, exc_info=True)
        except Exception as ex:
            logger.warning(ex, exc_info=True)
        finally:
            conn.close()
        if x509 is not None:
            logger.debug(f'negotiated protocol {negotiated_protocol} cipher {negotiated_cipher}')
            break
    
    if len(client_ca_list) > 0:
        client_ca_list = [cca.get_components() for cca in client_ca_list]

    return x509, certificate_chain, client_ca_list, negotiated_protocol, negotiated_cipher

def is_self_signed(cert :Certificate) -> bool:
    certificate_is_self_signed = False
    authority_key_identifier = None
    subject_key_identifier = None
    try:
        authority_key_identifier = hexlify(cert.extensions.get_extension_for_class(extensions.AuthorityKeyIdentifier).value.key_identifier).decode('utf-8')
    except extensions.ExtensionNotFound as ex:
        logger.debug(ex, exc_info=True)
        certificate_is_self_signed = True
    try:
        subject_key_identifier = hexlify(cert.extensions.get_extension_for_class(extensions.SubjectKeyIdentifier).value.digest).decode('utf-8')
    except extensions.ExtensionNotFound as ex:
        logger.debug(ex, exc_info=True)
        certificate_is_self_signed = True
    if subject_key_identifier == authority_key_identifier:
        certificate_is_self_signed = True
    return certificate_is_self_signed

def get_san(cert :Certificate) -> list:
    san = []
    try:
        san = cert.extensions.get_extension_for_class(SubjectAlternativeName).value.get_values_for_type(DNSName)
    except extensions.ExtensionNotFound as ex:
        logger.debug(ex, exc_info=True)
    return san

def get_basic_constraints(cert :Certificate) -> tuple[bool, int]:
    basic_constraints = None
    try:
        basic_constraints = cert.extensions.get_extension_for_class(extensions.BasicConstraints).value
    except extensions.ExtensionNotFound as ex:
        logger.debug(ex, exc_info=True)
    if not isinstance(basic_constraints, extensions.BasicConstraints):
        return None, None
    return basic_constraints.ca, basic_constraints.path_length

def key_usage_exists(cert :Certificate, key :str) -> bool:
    key_usage = None
    ext_key_usage = None
    try:
        key_usage = cert.extensions.get_extension_for_class(extensions.KeyUsage).value
    except extensions.ExtensionNotFound as ex:
        logger.debug(ex, exc_info=True)
    try:
        ext_key_usage = cert.extensions.get_extension_for_class(extensions.ExtendedKeyUsage).value
    except extensions.ExtensionNotFound as ex:
        logger.debug(ex, exc_info=True)
    if key_usage is None and ext_key_usage is None:
        logger.warning('no key usages could not be found')
        return False
    if isinstance(key_usage, extensions.KeyUsage) and hasattr(key_usage, key) and getattr(key_usage, key) is True:
        return True
    if isinstance(ext_key_usage, extensions.ExtendedKeyUsage) and key in [usage._name for usage in ext_key_usage if hasattr(usage, '_name')]:
        return True
    return False

def gather_key_usages(cert :Certificate) -> tuple[list, list, list]:
    certificate_extensions = []
    validator_key_usage = []
    validator_extended_key_usage = []
    for ext in cert.extensions:
        data = {
            'critical': ext.critical,
            'name': ext.oid._name # pylint: disable=protected-access
        }
        if isinstance(ext.value, extensions.UnrecognizedExtension):
            continue
        if isinstance(ext.value, extensions.CRLNumber):
            data[data['name']] = ext.value.crl_number
        if isinstance(ext.value, extensions.AuthorityKeyIdentifier):
            data[data['name']] = hexlify(ext.value.key_identifier).decode('utf-8')
            data['authority_cert_issuer'] = ', '.join([x.value for x in ext.value.authority_cert_issuer or []])
            data['authority_cert_serial_number'] = ext.value.authority_cert_serial_number
        if isinstance(ext.value, extensions.SubjectKeyIdentifier):
            data[data['name']] = hexlify(ext.value.digest).decode('utf-8')
        if isinstance(ext.value, (extensions.AuthorityInformationAccess, extensions.SubjectInformationAccess)):
            data[data['name']] = []
            for description in ext.value:
                data[data['name']].append({
                    'access_location': description.access_location.value,
                    'access_method': description.access_method._name, # pylint: disable=protected-access
                })
        if isinstance(ext.value, extensions.BasicConstraints):
            data['ca'] = ext.value.ca
            data['path_length'] = ext.value.path_length
        if isinstance(ext.value, extensions.DeltaCRLIndicator):
            data[data['name']] = ext.value.crl_number
        if isinstance(ext.value, (extensions.CRLDistributionPoints, extensions.FreshestCRL)):
            data[data['name']] = []
            for distribution_point in ext.value:
                data[data['name']].append({
                    'full_name': ', '.join([x.value for x in distribution_point.full_name or []]),
                    'relative_name': distribution_point.relative_name,
                    'reasons': distribution_point.reasons,
                    'crl_issuer': ', '.join([x.value for x in distribution_point.crl_issuer or []]),
                })
        if isinstance(ext.value, extensions.PolicyConstraints):
            data['policy_information'] = []
            data['user_notices'] = []
            for info in ext.value:
                if hasattr(info, 'require_explicit_policy'):
                    data['policy_information'].append({
                        'require_explicit_policy': info.require_explicit_policy,
                        'inhibit_policy_mapping': info.inhibit_policy_mapping,
                    })
                if hasattr(info, 'notice_reference'):
                    data['user_notices'].append({
                        'organization': info.notice_reference.organization,
                        'notice_numbers': info.notice_reference.notice_numbers,
                        'explicit_text': info.explicit_text,
                    })
        if isinstance(ext.value, extensions.ExtendedKeyUsage):
            data[data['name']] = [x._name for x in ext.value or []] # pylint: disable=protected-access
            if 'serverAuth' in data[data['name']]:
                validator_extended_key_usage.append('server_auth')
        if isinstance(ext.value, extensions.TLSFeature):
            data[data['name']] = []
            for feature in ext.value:
                if feature.value == 5:
                    data[data['name']].append('OCSP Must-Staple (rfc6066)')
                    validator_extended_key_usage.append('ocsp_signing')
                if feature.value == 17:
                    data[data['name']].append('multiple OCSP responses (rfc6961)')
                    validator_extended_key_usage.append('ocsp_signing')
        if isinstance(ext.value, extensions.InhibitAnyPolicy):
            data[data['name']] = ext.value.skip_certs
        if isinstance(ext.value, extensions.KeyUsage):
            data[data['name']] = []
            data['digital_signature'] = ext.value.digital_signature
            if ext.value.digital_signature:
                data[data['name']].append('digital_signature')
                validator_key_usage.append('digital_signature')
            data['content_commitment'] = ext.value.content_commitment
            if ext.value.content_commitment:
                data[data['name']].append('content_commitment')
                validator_key_usage.append('content_commitment')
            data['key_encipherment'] = ext.value.key_encipherment
            if ext.value.key_encipherment:
                data[data['name']].append('key_encipherment')
                validator_key_usage.append('key_encipherment')
            data['data_encipherment'] = ext.value.data_encipherment
            if ext.value.data_encipherment:
                data[data['name']].append('data_encipherment')
                validator_key_usage.append('data_encipherment')
            data['key_agreement'] = ext.value.key_agreement
            if ext.value.key_agreement:
                data[data['name']].append('key_agreement')
                validator_key_usage.append('key_agreement')
                data['decipher_only'] = ext.value.decipher_only
                if ext.value.decipher_only:
                    data[data['name']].append('decipher_only')
                    validator_key_usage.append('decipher_only')
                data['encipher_only'] = ext.value.encipher_only
                if ext.value.encipher_only:
                    data[data['name']].append('encipher_only')
                    validator_key_usage.append('encipher_only')
            data['key_cert_sign'] = ext.value.key_cert_sign
            if ext.value.key_cert_sign:
                data[data['name']].append('key_cert_sign')
                validator_key_usage.append('key_cert_sign')
            data['crl_sign'] = ext.value.crl_sign
            if ext.value.crl_sign:
                data[data['name']].append('crl_sign')
                validator_key_usage.append('crl_sign')
        if isinstance(ext.value, extensions.NameConstraints):
            data['permitted_subtrees'] = [x.value for x in ext.value.permitted_subtrees or []]
            data['excluded_subtrees'] = [x.value for x in ext.value.excluded_subtrees or []]
        if isinstance(ext.value, extensions.SubjectAlternativeName):
            data[data['name']] = [x.value for x in ext.value or []]
        if isinstance(ext.value, extensions.IssuerAlternativeName):
            data[data['name']] = [x.value for x in ext.value or []]
        if isinstance(ext.value, extensions.CertificateIssuer):
            data[data['name']] = [x.value for x in ext.value or []]
        if isinstance(ext.value, extensions.CRLReason):
            data[data['name']] = ext.value.reason
        if isinstance(ext.value, extensions.InvalidityDate):
            data[data['name']] = ext.value.invalidity_date
        if isinstance(ext.value, (extensions.PrecertificateSignedCertificateTimestamps, extensions.SignedCertificateTimestamps)):
            data[data['name']] = []
            for signed_cert_timestamp in ext.value:
                data[data['name']].append({
                    'version': signed_cert_timestamp.version.name,
                    'log_id': hexlify(signed_cert_timestamp.log_id).decode('utf-8'),
                    'timestamp': signed_cert_timestamp.timestamp,
                    'pre_certificate': signed_cert_timestamp.entry_type.value == 1,
                })
        if isinstance(ext.value, extensions.OCSPNonce):
            data[data['name']] = ext.value.nonce
        if isinstance(ext.value, extensions.IssuingDistributionPoint):
            data['full_name'] = ext.value.full_name
            data['relative_name'] = ext.value.relative_name
            data['only_contains_user_certs'] = ext.value.only_contains_user_certs
            data['only_contains_ca_certs'] = ext.value.only_contains_ca_certs
            data['only_some_reasons'] = ext.value.only_some_reasons
            data['indirect_crl'] = ext.value.indirect_crl
            data['only_contains_attribute_certs'] = ext.value.only_contains_attribute_certs
        certificate_extensions.append(data)
    return certificate_extensions, validator_key_usage, validator_extended_key_usage

def extract_certificate_common_name(cert :Certificate):
    for fields in cert.subject:
        current = str(fields.oid)
        if "commonName" in current:
            return fields.value
    return None

def validate_common_name(common_name :str, host :str) -> bool:
    if not isinstance(common_name, str):
        raise ValueError("invalid certificate_common_name provided")
    if not isinstance(host, str):
        raise ValueError("invalid host provided")
    if validators.domain(host) is not True:
        raise ValueError(f"provided an invalid domain {host}")
    if common_name.startswith('*.'):
        common_name_suffix = common_name.replace('*.', '')
        if validators.domain(common_name_suffix) is not True:
            return False
        return common_name_suffix == host or host.endswith(common_name_suffix)
    return validators.domain(common_name) is True

def match_hostname(host :str, cert :Certificate) -> bool:
    if not isinstance(host, str):
        raise ValueError("invalid host provided")
    if not isinstance(cert, Certificate):
        raise ValueError("invalid Certificate provided")
    if validators.domain(host) is not True:
        raise ValueError(f"provided an invalid domain {host}")
    certificate_san = []
    try:
        certificate_san = cert.extensions.get_extension_for_class(x509.SubjectAlternativeName).value.get_values_for_type(x509.DNSName)
    except extensions.ExtensionNotFound as ex:
        logger.debug(ex, exc_info=True)
    valid_common_name = False
    wildcard_hosts = set()
    domains = set()
    for fields in cert.subject:
        current = str(fields.oid)
        if "commonName" in current:
            valid_common_name = validate_common_name(fields.value, host)
    for san in certificate_san:
        if san.startswith('*.'):
            wildcard_hosts.add(san)
        else:
            domains.add(san)
    matched_wildcard = False
    for wildcard in wildcard_hosts:
        check = wildcard.replace('*', '')
        if host.endswith(check):
            matched_wildcard = True
            break

    return valid_common_name is True and (matched_wildcard is True or host in domains)

def validate_certificate_chain(der :bytes, pem_certificate_chain :list, validator_key_usage :list, validator_extended_key_usage :list):
    # TODO perhaps remove certvalidator, consider once merged: https://github.com/pyca/cryptography/issues/2381
    ctx = ValidationContext(allow_fetching=True, revocation_mode='hard-fail', weak_hash_algos=set(["md2", "md5", "sha1"]))
    validator = CertificateValidator(der, validation_context=ctx, intermediate_certs=pem_certificate_chain)
    return validator.validate_usage(
        key_usage=set(validator_key_usage),
        extended_key_usage=set(validator_extended_key_usage),
    )



import os
import pathlib
from setuptools import setup, find_packages

setup(
    name="tls-verify",
    version="0.1.7",
    author='Christopher Langton',
    author_email='chris@langton.cloud',
    description="Validate the security of your TLS connections so that they deserve your trust.",
    long_description=pathlib.Path(os.path.join(os.path.dirname(__file__), "README.md")).read_text(),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/chrislangton/py-tls-veryify",
    project_urls={
        "Source": "https://gitlab.com/chrislangton/py-tls-veryify",
        "Documentation": "https://gitlab.com/chrislangton/py-tls-veryify/-/blob/main/README.md",
        "Tracker": "https://gitlab.com/chrislangton/py-tls-veryify/-/issues",
    },
    classifiers=[
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    include_package_data=True,
    install_requires=[
        'certifi==2021.5.30',
        'cryptography==35.0.0',
        'asn1crypto==1.4.0',
        'certvalidator==0.11.1',
        'oscrypto==1.2.1',
        'pyOpenSSL==21.0.0',
        'validators==0.18.2',
        'idna==3.2',
        'tabulate==0.8.9'
    ],
    entry_points = {
        'console_scripts': ['tlsverify=tlsverify.cli:cli'],
    },
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.8",
    options={"bdist_wheel": {"universal": "1"}},
)
